-----------------------------------------------------------------------
-- To do and/or ideas
--
-- From Rabbit:
-- * Parse damage messages from swstats and damagemeters.
-- * Automatically output our damage done and position to chat when a boss has
-- been killed; this is easy with Babble-Boss, CHAT_MSG_HOSTILE_DEATH and
-- UnitClassification.
-- * Maybe add an auto-reset for boss kill as well.
--
-- From Nymbia:
-- * Pick up the slack of users without the addon by splitting up the members
-- without the addon among those with it.  ie, in a 25 man raid where 10 users
-- have Violation, each of them is assigned to parse the damage of either 1 or
-- 2 other raid members.
-- * Possibly keep two running totals:  'this fight' and 'overall', to allow
-- for statistics to be displayed for a single encounter without requiring
-- constant resets.
--
-- From <person>:
-- * Brew my coffee when I win the damage meter contest after a boss fight.
--
-----------------------------------------------------------------------

-----------------------------------------------------------------------
-- Locale
-----------------------------------------------------------------------
local L = AceLibrary("AceLocale-2.2"):new("Violation")

L:RegisterTranslations("enUS", function() return {
	["Track"] = true,
	["What data types to track."] = true,

	["Resetting data as requested by %s."] = true,
	["Reset"] = true,
	["Options for when to reset the data."] = true,
	["Reset now"] = true,
	["Immediately resets your data."] = true,
	["Reset group"] = true,
	["Resets the Violation data for every group member. Only the group leader or an officer can do this."] = true,
	["Leader resets"] = true,
	["Accept reset requests from raid leaders or officers."] = true,
	["Auto-reset"] = true,
	["Entering combat"] = true,
	["Automatically reset the data when entering combat."] = true,
	["Leaving combat"] = true,
	["Automatically reset the data when leaving combat."] = true,
	["Print on reset"] = true,
	["Prints your current data locally when your data is reset."] = true,

	["Text"] = true,
	["What data to show on the FuBar panel, if available."] = true,

	["Toggle whether to track %q."] = true,
	["Toggle whether to display data from %q on the FuBar panel."] = true,

	["No data gathered yet."] = true,
	["|cffeda55fClick|r to reset."] = true,
} end)

-----------------------------------------------------------------------
-- Locals
-----------------------------------------------------------------------
local tablet = AceLibrary("Tablet-2.0")
local BC = AceLibrary("Babble-Class-2.2")
local roster = AceLibrary("Roster-2.1")

local coloredNames = setmetatable({}, {__index =
	function(self, key)
		local obj = roster:GetUnitObjectFromName(key)
		if not obj then return nil end
		self[key] = "|cff" .. BC:GetHexColor(obj.class) .. key .. "|r"
		return self[key]
	end
})

local db = nil

local _G = getfenv(0)

local red = "Interface\\AddOns\\Violation\\icons\\red"
local blue = "Interface\\AddOns\\Violation\\icons\\blue"
local grey = "Interface\\AddOns\\Violation\\icons\\grey"

-----------------------------------------------------------------------
-- Local heap
-----------------------------------------------------------------------

local new, del
do
	local cache = setmetatable({},{__mode='k'})
	function new(...)
		local t = next(cache)
		if t then
			cache[t] = nil
			for i = 1, select("#", ...) do
				table.insert(t, (select(i, ...)))
			end
			return t
		else
			return {...}
		end
	end
	function del(t)
		if type(t) == "table" then
			for k, v in pairs(t) do
				if type(v) == "table" then
					del(v)
				end
				t[k] = nil
			end
			cache[t] = true
		end
		return nil
	end
end

-----------------------------------------------------------------------
-- Options
-----------------------------------------------------------------------

local options = {
	type = "group",
	args = {
		track = {
			type = "group",
			name = L["Track"],
			desc = L["What data types to track."],
			disabled = function() return not Violation:IsActive() end,
			args = {}, -- Auto-populated by module toggles.
			order = 101,
		},
		text = {
			type = "group",
			name = L["Text"],
			desc = L["What data to show on the FuBar panel, if available."],
			disabled = function() return not FuBar end,
			args = {}, -- Auto-populated by module toggles.
			order = 102,
		},
		spacer = {
			type = "header",
			name = " ",
			order = 150,
		},
		print = {
			type = "toggle",
			name = L["Print on reset"],
			desc = L["Prints your current data locally when your data is reset."],
			get = function() return Violation.db.profile.printOnReset end,
			set = function(v) Violation.db.profile.printOnReset = v end,
			order = 200,
		},
		reset = {
			type = "group",
			name = L["Reset"],
			desc = L["Options for when to reset the data."],
			args = {
				now = {
					type = "execute",
					name = L["Reset now"],
					desc = L["Immediately resets your data."],
					func = function() Violation:Reset() end,
					order = 100,
				},
				group = {
					type = "execute",
					name = L["Reset group"],
					desc = L["Resets the Violation data for every group member. Only the group leader or an officer can do this."],
					func = function() Violation:ResetGroup() end,
					disabled = function()
						return not IsRaidLeader() and not IsRaidOfficer() and not IsPartyLeader()
					end,
					order = 101,
				},
				leader = {
					type = "toggle",
					name = L["Leader resets"],
					desc = L["Accept reset requests from raid leaders or officers."],
					get = function() return Violation.db.profile.leaderReset end,
					set = function(v) Violation.db.profile.leaderReset = v end,
					order = 102,
				},
				spacer = {
					type = "header",
					name = " ",
					order = 150,
				},
				autoheader = {
					type = "header",
					name = L["Auto-reset"],
					order = 200,
				},
				gaincombat = {
					type = "toggle",
					name = L["Entering combat"],
					desc = L["Automatically reset the data when entering combat."],
					get = function() return Violation.db.profile.gainCombat end,
					set = function(v)
						if v then
							Violation:RegisterEvent("PLAYER_REGEN_DISABLED", "Reset")
						else
							Violation:UnregisterEvent("PLAYER_REGEN_DISABLED")
						end
						Violation.db.profile.gainCombat = v
					end,
					order = 201,
				},
				leavecombat = {
					type = "toggle",
					name = L["Leaving combat"],
					desc = L["Automatically reset the data when leaving combat."],
					get = function() return Violation.db.profile.leaveCombat end,
					set = function(v)
						if v then
							Violation:RegisterEvent("PLAYER_REGEN_ENABLED", "Reset")
						else
							Violation:UnregisterEvent("PLAYER_REGEN_ENABLED")
						end
						Violation.db.profile.leaveCombat = v
					end,
					order = 202,
				},
			},
			order = 201,
		},
	},
}

-----------------------------------------------------------------------
-- Addon declaration
-----------------------------------------------------------------------
Violation = AceLibrary("AceAddon-2.0"):new("AceEvent-2.0", "AceDB-2.0", "AceConsole-2.0", "AceComm-2.0", "FuBarPlugin-2.0", "AceModuleCore-2.0", "Parser-3.0")
local Violation = Violation

Violation:SetModuleMixins("AceEvent-2.0")
Violation.hasIcon = true

-----------------------------------------------------------------------
-- Module prototype
-----------------------------------------------------------------------

Violation.modulePrototype.core = Violation
local registered = {}
function Violation.modulePrototype:OnInitialize()
	if registered[self.name] then
		error("A module has been initialized twice for some reason.")
		return
	end
	self.core:OnRegister(self)
	registered[self.name] = true
end

function Violation.modulePrototype:OnEnable(first)
	--[[if first then
		self.core:Print("First time we're enabling "..self.name..", returning.")
		self.core:ToggleModuleActive(self, false)]]
	if type(self.OnModuleEnable) == "function" then
		self:OnModuleEnable()
	end
end

function Violation.modulePrototype:GetColoredName(unit)
	return coloredNames[unit]
end

-----------------------------------------------------------------------
-- Implicit module interface
-----------------------------------------------------------------------

local AceOO = AceLibrary("AceOO-2.0")
local IViolation = AceOO.Interface{

	-- Optional methods:
	--
	-- -- Called when the module is enabled by the core. Any event registration
	-- -- should happen here.
	-- OnModuleEnable
	--
	-- -- Called when the module is initially registered.
	-- OnModuleRegister

	-- Must return the current module revision as a number.
	GetRevision = "function",

	-- Must return the module name as it should be displayed in the menu, for
	-- example "Damage done".
	GetDisplayName = "function",

	-- Must return whether the modules data is dirty or not. If it is, a
	-- broadcast will be scheduled by the core.
	IsDirty = "function",

	-- Must return the data that should be broadcasted to the other players.
	-- Must return 1 number only.
	GetBroadcastData = "function",

	-- Invoked when remote data is received for this module. Usually invoked by
	-- the core, but can also be invoked by other modules when they parse data
	-- from other addons, for example.
	-- Will only get two arguments, the first will be the person the data
	-- belongs to and the second will be a number.
	OnRemoteData = "function",

	-- Invoked by the core when all data should be reset.
	OnReset = "function",

	-- Must return a universal unique ID that will be used for this module when
	-- sending data (where the universe is Violation)
	-- (Not used now, we use a memoization of the module name instead.
	-- GetUUID = "function",

	-- Called when the FuBar tooltip is updated, you can add categories and
	-- lines to the FuBar tablet from here.
	OnTooltipUpdate = "function",

	-- The text that should be shown on the FuBar panel. Only invoked if this
	-- module should actually show text on the panel. Remember to include some
	-- prefix that identifies the data.
	GetDisplayText = "function",
}

-----------------------------------------------------------------------
-- Addon initialization
-----------------------------------------------------------------------

function Violation:OnInitialize()
	local revision = tonumber(("$Revision: 28690 $"):sub(12, -3)) or 1
	if not self.version then self.version = "1.0" end
	self.version = self.version .. "." .. revision
	self.revision = revision

	_G.VIOLATION_VERSION = self.version
	_G.VIOLATION_REVISION = self.revision

	self:RegisterDB("ViolationDB")
	self:RegisterDefaults("char", {
		enabledModules = {},
		textModules = {},
		gainCombat = nil,
		leaveCombat = nil,
		leaderReset = true,
		printOnReset = true,
	})

	self.OnMenuRequest = options
	self.hasNoColor = true

	self:RegisterChatCommand({ "/violation" }, self.OnMenuRequest, "VIOLATION")

	self:SetCommPrefix("Violation")
	self:SetDefaultCommPriority("BULK")

	self:SetIcon(grey)
end

function Violation:OnEnable(first)
	db = self.db.char

	if first then
		local memoizations = new()
		for name in self:IterateModules() do
			table.insert(memoizations, name)
		end
		self:RegisterMemoizations(memoizations)
		memoizations = del(memoizations)
	end

	-- Enable/disable all modules.
	for name, mod in self:IterateModules() do
		self:ToggleModuleActive(name, db.enabledModules[name])
	end

	self:Reset()

	self:RegisterComm("Violation", "GROUP", "OnComm")
	self:ScheduleRepeatingEvent("Broadcast", self.Broadcast, 5, self)

	if db.gainCombat then
		self:RegisterEvent("PLAYER_REGEN_DISABLED", "Reset")
	end

	if db.leaveCombat then
		self:RegisterEvent("PLAYER_REGEN_ENABLED", "Reset")
	end

	self:SetIcon(red)
end

function Violation:OnDisable()
	-- Disable all modules.
	for name, mod in self:IterateModules() do
		self:ToggleModuleActive(name, false)
	end

	self:SetIcon(grey)
end

function Violation:Reset()
	for name, mod in self:IterateModules() do
		if self:IsModuleActive(name) then
			if db.printOnReset then
				local text = mod:GetDisplayText()
				if text then self:Print(text) end
			end
			mod:OnReset()
		end
	end

	self:UpdateDisplay()
end

-----------------------------------------------------------------------
-- Module handling
-----------------------------------------------------------------------

function Violation:OnRegister(mod)
	if not AceOO.inherits(mod, IViolation) then
		error(string.format("The module %q does not inherit from the IViolation interface.", tostring(mod.name)))
	end

	local moduleName = mod.name
	local displayName = mod:GetDisplayName()

	options.args.track.args[moduleName] = {
		type = "toggle",
		name = displayName,
		desc = string.format(L["Toggle whether to track %q."], displayName),
		get = function()
			return self.db.char.enabledModules[moduleName]
		end,
		set = function(v)
			self.db.char.enabledModules[moduleName] = v
			self:ToggleModuleActive(moduleName, v)
			self:UpdateDisplay()
		end
	}

	options.args.text.args[moduleName] = {
		type = "toggle",
		name = displayName,
		desc = string.format(L["Toggle whether to display data from %q on the FuBar panel."], displayName),
		get = function()
			return self.db.char.textModules[moduleName]
		end,
		set = function(v)
			self.db.char.textModules[moduleName] = v
			self:UpdateText()
		end
	}

	if type(mod.OnModuleRegister) == "function" then
		mod:OnModuleRegister()
	end
end

-----------------------------------------------------------------------
-- UI
-----------------------------------------------------------------------

function Violation:OnClick()
	self:Reset()
end

function Violation:OnTooltipUpdate()
	for name, mod in self:IterateModules() do
		if self:IsModuleActive(name) then
			mod:OnTooltipUpdate()
		end
	end

	tablet:SetHint(L["|cffeda55fClick|r to reset."])
end

function Violation:OnTextUpdate()
	local hasText = nil
	for k, v in pairs(db.textModules) do
		if v then
			local mod = self:GetModule(k)
			assert(mod, "Some module doesn't exist.")
			local data = mod:GetDisplayText()
			if data then
				if not hasText then hasText = new(data)
				else table.insert(hasText, data) end
			end
		end
	end

	if not hasText then
		self:SetText(self.name)
	else
		self:SetText(unpack(hasText))
	end
end

-----------------------------------------------------------------------
-- Data broadcasting and recieving
-----------------------------------------------------------------------
function Violation:Broadcast()
	local tbl = nil
	for name, mod in self:IterateModules() do
		if self:IsModuleActive(name) and mod:IsDirty() then
			if not tbl then tbl = new() end
			table.insert(tbl, name)
			table.insert(tbl, mod:GetBroadcastData())
		end
	end
	if tbl then
		self:SendCommMessage("GROUP", "B", unpack(tbl))
	end
	tbl = del(tbl)
end

function Violation:ResetGroup()
	if not IsRaidLeader() and not IsRaidOfficer() and not IsPartyLeader() then
		error("Only the group leader or an officer can do that.")
		return
	end

	self:SendCommMessage("GROUP", "LeaderReset")
end

function Violation:OnComm(prefix, sender, distribution, command, ...)
	if distribution ~= "GROUP" or type(command) ~= "string" then return end

	if command == "LeaderReset" then
		local obj = roster:GetUnitObjectFromName(sender)
		if obj.rank > 0 then
			self:Print(string.format(L["Resetting data as requested by %s."], sender))
			self:Reset()
		end
	elseif command == "B" then
		local total = select("#", ...)
		local i
		for i = 1, total do
			local name = select(i, ...)
			if self:HasModule(name) and self:IsModuleActive(name) then
				self:GetModule(name):OnRemoteData(sender, select(i + 1, ...))
			end
			i = i + 1
			if i > total then break end
		end
	end
end

